"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const typescript_1 = __importDefault(require("typescript"));
const createRule_1 = require("../utils/createRule");
const getParserServices_1 = require("../utils/getParserServices");
const loc_1 = require("../utils/loc");
const snapshot_1 = require("../utils/snapshot");
const messages = {
    TypeScriptCompileError: 'TypeScript compile error: {{ message }}',
    FileIsNotIncludedInTsconfig: 'Expected to find a file "{{ fileName }}" present.',
    TypesDoNotMatch: 'Expected type to be: {{ expected }}, got: {{ actual }}',
    OrphanAssertion: 'Can not match a node to this assertion.',
    Multiple$ExpectTypeAssertions: 'This line has 2 or more $ExpectType assertions.',
    ExpectedErrorNotFound: 'Expected an error on this line, but found none.',
    TypeSnapshotNotFound: 'Type Snapshot not found. Please consider running ESLint in FIX mode: eslint --fix',
    TypeSnapshotDoNotMatch: 'Expected type from Snapshot to be: {{ expected }}, got: {{ actual }}',
    SyntaxError: 'Syntax Error: {{ message }}',
};
// The default options for the rule.
const defaultOptions = {
    // expectError: true,
    // expectType: true,
    // expectTypeSnapshot: true,
    disableExpectTypeSnapshotFix: false,
};
// The schema for the rule options.
const schema = [
    {
        type: 'object',
        properties: {
            // expectError: {
            //   type: 'boolean',
            // },
            // expectType: {
            //   type: 'boolean',
            // },
            // expectTypeSnapshot: {
            //   type: 'boolean',
            // },
            disableExpectTypeSnapshotFix: {
                type: 'boolean',
            },
        },
        additionalProperties: false,
    },
];
exports.name = 'expect';
exports.rule = createRule_1.createRule({
    name: exports.name,
    meta: {
        type: 'problem',
        docs: {
            description: 'Expects type error, type snapshot or type.',
            category: 'Possible Errors',
            recommended: 'error',
            requiresTypeChecking: true,
        },
        fixable: 'code',
        schema,
        messages,
    },
    defaultOptions: [defaultOptions],
    create(context, [options]) {
        validate(context, options);
        return {};
    },
});
function validate(context, options) {
    const parserServices = getParserServices_1.getParserServices(context);
    const { program } = parserServices;
    const fileName = context.getFilename();
    const sourceFile = program.getSourceFile(fileName);
    if (!sourceFile) {
        context.report({
            loc: {
                line: 1,
                column: 0,
            },
            messageId: 'FileIsNotIncludedInTsconfig',
            data: {
                fileName,
            },
        });
        return;
    }
    const checker = program.getTypeChecker();
    // Don't care about emit errors.
    const diagnostics = typescript_1.default.getPreEmitDiagnostics(program, sourceFile);
    if (sourceFile.isDeclarationFile || !/\$Expect(Type|Error)/.test(sourceFile.text)) {
        // Normal file.
        for (const diagnostic of diagnostics) {
            addDiagnosticFailure(diagnostic);
        }
        return;
    }
    const { errorLines, typeAssertions, duplicates, syntaxErrors } = parseAssertions(sourceFile);
    for (const line of duplicates) {
        context.report({
            messageId: 'Multiple$ExpectTypeAssertions',
            loc: {
                line: line + 1,
                column: 0,
            },
        });
    }
    const seenDiagnosticsOnLine = new Set();
    for (const diagnostic of diagnostics) {
        const line = lineOfPosition(diagnostic.start, sourceFile);
        seenDiagnosticsOnLine.add(line);
        if (!errorLines.has(line)) {
            addDiagnosticFailure(diagnostic);
        }
    }
    for (const line of errorLines) {
        if (!seenDiagnosticsOnLine.has(line)) {
            context.report({
                messageId: 'ExpectedErrorNotFound',
                loc: {
                    line: line + 1,
                    column: 0,
                },
            });
        }
    }
    for (const { type, line } of syntaxErrors) {
        context.report({
            messageId: 'SyntaxError',
            data: {
                message: type === 'MissingExpectType'
                    ? '$ExpectType requires type argument (e.g. // $ExpectType "string")'
                    : '$ExpectTypeSnapshot requires snapshot name argument (e.g. // $ExpectTypeSnapshot MainComponentAPI)',
            },
            loc: {
                line: line + 1,
                column: 0,
            },
        });
    }
    for (const [_line, assertion] of typeAssertions) {
        if (assertion.assertionType === 'snapshot') {
            assertion.expected = snapshot_1.getTypeSnapshot(fileName, assertion.snapshotName);
        }
    }
    const { unmetExpectations, unusedAssertions } = getExpectTypeFailures(sourceFile, typeAssertions, checker);
    for (const { node, assertion, actual } of unmetExpectations) {
        const templateDescriptor = {
            data: {
                expected: assertion.expected,
                actual,
            },
            loc: loc_1.loc(sourceFile, node),
        };
        if (assertion.assertionType === 'snapshot') {
            const { snapshotName } = assertion;
            const start = node.getStart();
            const fix = () => {
                let applied = false;
                return {
                    range: [start, start],
                    // Trick: Function `fix` is ALWAYS run, and result is collected. However RuleFix objects are only read if `--fix` is passed
                    get text() {
                        if (!applied) {
                            // Make sure we update snapshot only on first read of this object
                            applied = true;
                            if (!options.disableExpectTypeSnapshotFix) {
                                snapshot_1.updateTypeSnapshot(fileName, snapshotName, actual);
                            }
                        }
                        return '';
                    },
                };
            };
            if (typeof assertion.expected === 'undefined') {
                context.report(Object.assign(Object.assign({}, templateDescriptor), { messageId: 'TypeSnapshotNotFound', fix }));
            }
            else {
                context.report(Object.assign(Object.assign({}, templateDescriptor), { messageId: 'TypeSnapshotDoNotMatch', fix }));
            }
        }
        else {
            context.report(Object.assign(Object.assign({}, templateDescriptor), { messageId: 'TypesDoNotMatch' }));
        }
    }
    for (const line of unusedAssertions) {
        context.report({
            messageId: 'OrphanAssertion',
            loc: {
                line: line + 1,
                column: 0,
            },
        });
    }
    function addDiagnosticFailure(diagnostic) {
        if (diagnostic.file === sourceFile) {
            const message = `${typescript_1.default.flattenDiagnosticMessageText(diagnostic.messageText, '\n')}`;
            context.report({
                messageId: 'TypeScriptCompileError',
                data: {
                    message,
                },
                loc: {
                    line: diagnostic.start + 1,
                    column: diagnostic.length,
                },
            });
        }
        else {
            context.report({
                messageId: 'TypeScriptCompileError',
                data: {
                    message: `${fileName}${diagnostic.messageText}`,
                },
                loc: {
                    line: 1,
                    column: 0,
                },
            });
        }
    }
}
function parseAssertions(sourceFile) {
    const errorLines = new Set();
    const typeAssertions = new Map();
    const duplicates = [];
    const syntaxErrors = [];
    const { text } = sourceFile;
    const commentRegexp = /\/\/(.*)/g;
    const lineStarts = sourceFile.getLineStarts();
    let curLine = 0;
    while (true) {
        const commentMatch = commentRegexp.exec(text);
        if (commentMatch === null) {
            break;
        }
        // Match on the contents of that comment so we do nothing in a commented-out assertion,
        // i.e. `// foo; // $ExpectType number`
        const match = /^ \$Expect(TypeSnapshot|Type|Error)( (.*))?$/.exec(commentMatch[1]);
        if (match === null) {
            continue;
        }
        const line = getLine(commentMatch.index);
        switch (match[1]) {
            case 'TypeSnapshot':
                const snapshotName = match[3];
                if (snapshotName) {
                    if (typeAssertions.delete(line)) {
                        duplicates.push(line);
                    }
                    else {
                        typeAssertions.set(line, {
                            assertionType: 'snapshot',
                            snapshotName,
                        });
                    }
                }
                else {
                    syntaxErrors.push({
                        type: 'MissingSnapshotName',
                        line,
                    });
                }
                break;
            case 'Error':
                if (errorLines.has(line)) {
                    duplicates.push(line);
                }
                errorLines.add(line);
                break;
            case 'Type':
                const expected = match[3];
                if (expected) {
                    // Don't bother with the assertion if there are 2 assertions on 1 line. Just fail for the duplicate.
                    if (typeAssertions.delete(line)) {
                        duplicates.push(line);
                    }
                    else {
                        typeAssertions.set(line, { assertionType: 'manual', expected });
                    }
                }
                else {
                    syntaxErrors.push({
                        type: 'MissingExpectType',
                        line,
                    });
                }
                break;
        }
    }
    return { errorLines, typeAssertions, duplicates, syntaxErrors };
    function getLine(pos) {
        // advance curLine to be the line preceding 'pos'
        while (lineStarts[curLine + 1] <= pos) {
            curLine++;
        }
        // If this is the first token on the line, it applies to the next line.
        // Otherwise, it applies to the text to the left of it.
        return isFirstOnLine(text, lineStarts[curLine], pos) ? curLine + 1 : curLine;
    }
}
function isFirstOnLine(text, lineStart, pos) {
    for (let i = lineStart; i < pos; i++) {
        if (text[i] !== ' ') {
            return false;
        }
    }
    return true;
}
function matchReadonlyArray(actual, expected) {
    if (!(/\breadonly\b/.test(actual) && /\bReadonlyArray\b/.test(expected)))
        return false;
    const readonlyArrayRegExp = /\bReadonlyArray</y;
    const readonlyModifierRegExp = /\breadonly /y;
    // A<ReadonlyArray<B<ReadonlyArray<C>>>>
    // A<readonly B<readonly C[]>[]>
    let expectedPos = 0;
    let actualPos = 0;
    let depth = 0;
    while (expectedPos < expected.length && actualPos < actual.length) {
        const expectedChar = expected.charAt(expectedPos);
        const actualChar = actual.charAt(actualPos);
        if (expectedChar === actualChar) {
            expectedPos++;
            actualPos++;
            continue;
        }
        // check for end of readonly array
        if (depth > 0 &&
            expectedChar === '>' &&
            actualChar === '[' &&
            actualPos < actual.length - 1 &&
            actual.charAt(actualPos + 1) === ']') {
            depth--;
            expectedPos++;
            actualPos += 2;
            continue;
        }
        // check for start of readonly array
        readonlyArrayRegExp.lastIndex = expectedPos;
        readonlyModifierRegExp.lastIndex = actualPos;
        if (readonlyArrayRegExp.test(expected) && readonlyModifierRegExp.test(actual)) {
            depth++;
            expectedPos += 14; // "ReadonlyArray<".length;
            actualPos += 9; // "readonly ".length;
            continue;
        }
        return false;
    }
    return true;
}
function getExpectTypeFailures(sourceFile, typeAssertions, checker) {
    const unmetExpectations = [];
    // Match assertions to the first node that appears on the line they apply to.
    // `forEachChild` isn't available as a method in older TypeScript versions, so must use `ts.forEachChild` instead.
    typescript_1.default.forEachChild(sourceFile, function iterate(node) {
        const line = lineOfPosition(node.getStart(sourceFile), sourceFile);
        const assertion = typeAssertions.get(line);
        if (assertion !== undefined) {
            const { expected } = assertion;
            // https://github.com/Microsoft/TypeScript/issues/14077
            if (node.kind === typescript_1.default.SyntaxKind.ExpressionStatement) {
                node = node.expression;
            }
            const type = checker.getTypeAtLocation(getNodeForExpectType(node));
            const actual = type
                ? checker.typeToString(type, /*enclosingDeclaration*/ undefined, typescript_1.default.TypeFormatFlags.NoTruncation)
                : '';
            if (!expected || (actual !== expected && !matchReadonlyArray(actual, expected))) {
                unmetExpectations.push({ assertion, node, actual });
            }
            typeAssertions.delete(line);
        }
        typescript_1.default.forEachChild(node, iterate);
    });
    return { unmetExpectations, unusedAssertions: typeAssertions.keys() };
}
function getNodeForExpectType(node) {
    if (node.kind === typescript_1.default.SyntaxKind.VariableStatement) {
        // ts2.0 doesn't have `isVariableStatement`
        const { declarationList: { declarations }, } = node;
        if (declarations.length === 1) {
            const { initializer } = declarations[0];
            if (initializer) {
                return initializer;
            }
        }
    }
    return node;
}
function lineOfPosition(pos, sourceFile) {
    return sourceFile.getLineAndCharacterOfPosition(pos).line;
}
